"""Command line interface for glyTrait."""

from pathlib import Path

import click
import emoji

from glytrait.api import GlyTrait
from glytrait.exception import GlyTraitError
from glytrait.formula import save_builtin_formula

UNDIFINED = "__UNDEFINED__"


def save_builtin_formulas_callback(ctx, param, value):
    """Save a copy of the built-in formulas."""
    if value == UNDIFINED:
        return
    if Path(value).exists() and not Path(value).is_dir():
        msg = "The path to save the built-in formulas must be a directory."
        raise click.BadParameter(msg)
    else:
        save_builtin_formula(value)
        msg = (
            f"Built-in formulas saved to: "
            f"{value}/struc_builtin_formulas.txt, {value}/comp_builtin_formulas.txt"
        )
        click.echo(emoji.emojize(msg))
        ctx.exit()


@click.command()
@click.argument(
    "abundance-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=False,
)
@click.argument(
    "glycan-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    required=False,
)
@click.option(
    "-g",
    "--group-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Group file path.",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    help="Output file path. Default is the input file name with '_glytrait.xlsx' "
    "suffix.",
)
@click.option(
    "-m",
    "--mode",
    type=click.Choice(["structure", "composition", "S", "C"]),
    default="structure",
    help="Mode to run glyTrait, either 'structure' or 'composition'. "
    "You can also use 'S' or 'C' for short. "
    "Default is 'structure'.",
)
@click.option(
    "-r",
    "--filter-glycan-ratio",
    type=click.FLOAT,
    default=1,
    help="Glycans with missing value ratio greater than this value will be filtered out.",
)
@click.option(
    "-i",
    "--impute-method",
    type=click.Choice(["zero", "min", "lod", "mean", "median"]),
    default="zero",
    help="Method to impute missing values.",
)
@click.option(
    "-l",
    "--sia-linkage",
    is_flag=True,
    help="Include sialic acid linkage traits.",
)
@click.option(
    "-f",
    "--formula-file",
    type=click.Path(exists=True),
    help="User formula file.",
)
@click.option(
    "--filter/--no-filter",
    default=True,
    help="Filter out invalid derived traits. Default is filtering."
    "Use --no-filter to disable filtering.",
)
@click.option(
    "-c",
    "--corr-threshold",
    type=click.FLOAT,
    default=1,
    help="Threshold for correlation between traits. "
    "Default is 1, which means only traits with perfect collinearity "
    "will be filtered. Use -1 to disable filtering.",
)
@click.option(
    "-b",
    "--builtin-formulas",
    type=click.Path(),
    callback=save_builtin_formulas_callback,
    is_eager=True,
    expose_value=False,
    default=UNDIFINED,
    help="The directory path to save a copy of the built-in formulas.",
)
@click.version_option()
def cli(
    abundance_file,
    glycan_file,
    group_file,
    output_dir,
    mode,
    filter_glycan_ratio,
    impute_method,
    sia_linkage,
    formula_file,
    filter,
    corr_threshold,
):
    """Run the glytrait workflow."""
    if abundance_file is None:
        msg = r"""
Welcome to GlyTrait!

   _____ _    _______        _ _   
  / ____| |  |__   __|      (_) |  
 | |  __| |_   _| |_ __ __ _ _| |_ 
 | | |_ | | | | | | '__/ _` | | __|
 | |__| | | |_| | | | | (_| | | |_ 
  \_____|_|\__, |_|_|  \__,_|_|\__|
            __/ |                  
           |___/                   

Use `glytrait --help` for more information.
"""
        click.echo(msg)
        return

    if output_dir is None:
        output_dir = Path(abundance_file).with_name(
            Path(abundance_file).stem + "_glytrait"
        )
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    mode = "composition" if mode.lower() in ["c", "composition"] else "structure"
    try:
        gt = GlyTrait(
            mode=mode,
            filter_max_na=filter_glycan_ratio,
            impute_method=impute_method,
            post_filtering=filter,
            correlation_threshold=corr_threshold,
            sia_linkage=sia_linkage,
            custom_formula_file=formula_file,
        )
        gt.run(
            output_dir=str(output_dir),
            abundance_file=abundance_file,
            glycan_file=glycan_file,
            group_file=group_file,
        )
    except GlyTraitError as e:
        raise click.UsageError(str(e) + emoji.emojize(" :thumbs_down:"))
    msg = f"Done :thumbs_up:! Output written to {output_dir}."
    click.echo(emoji.emojize(msg))


if __name__ == "__main__":
    cli()



import warnings

import pandas as pd
import pingouin as pg

from glytrait.data_type import DerivedTraitTable, GroupSeries

__all__ = ["t_test", "anova"]


def t_test(trait_df: DerivedTraitTable, groups: GroupSeries) -> pd.DataFrame:
    """Perform t-test for the trait data.

    Args:
        trait_df (DerivedTraitTable): Dataframe containing the trait data.
        groups (GroupSeries): Series containing the group information.

    Returns:
        pd.DataFrame: Dataframe containing the t-test results,
            with trait names as index.
    """
    group_names = groups.unique()
    assert len(group_names) == 2, "t-test requires two groups."

    group_data = trait_df.groupby(groups)
    group_1_data = group_data.get_group(group_names[0])
    group_2_data = group_data.get_group(group_names[1])

    results: list[pd.DataFrame] = []
    for trait_name in trait_df.columns:
        t_test_result = pg.ttest(group_1_data[trait_name], group_2_data[trait_name])
        t_test_result.index = [trait_name]
        results.append(t_test_result)
    result_df = pd.concat(results)
    result_df["p-val-adj"] = pg.multicomp(result_df["p-val"].values, method="fdr_bh")[1]
    return result_df


def anova(
    trait_df: DerivedTraitTable, groups: GroupSeries
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Perform ANOVA for the trait data.

    Args:
        trait_df (DerivedTraitTable): Dataframe containing the trait data.
        groups (GroupSeries): Series containing the group information.

    Returns:
        tuple[pd.DataFrame, pd.DataFrame]:
            Tuple containing ANOVA results and post-hoc results.
    """
    trait_names = trait_df.columns.tolist()
    prepared_df = _prepare_for_anova(trait_df, groups)
    anova_result_df = _anova(prepared_df, trait_names)
    sig_traits = anova_result_df[anova_result_df["reject"]].index.tolist()
    post_hoc_result_df = _post_hoc(prepared_df, sig_traits)
    return anova_result_df, post_hoc_result_df


def _prepare_for_anova(
    trait_df: DerivedTraitTable, groups: GroupSeries
) -> pd.DataFrame:
    """Combine the trait data and group information."""
    groups = groups.rename("group")
    return trait_df.merge(groups, left_index=True, right_index=True)


def _anova(prepared_df: pd.DataFrame, trait_names: list[str]) -> pd.DataFrame:
    """Perform ANOVA for the trait data.

    Args:
        prepared_df: Dataframe containing the trait data and group information.
        trait_names: List of trait names.
    """
    anove_results: list[pd.DataFrame] = []
    for trait in trait_names:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            anova_result = pg.welch_anova(data=prepared_df, dv=trait, between="group")
        anova_result.index = [trait]
        anove_results.append(anova_result)
    return _tidy_anova_result(pd.concat(anove_results))


def _tidy_anova_result(result_df: pd.DataFrame) -> pd.DataFrame:
    """Tidy the ANOVA result."""
    result_df = result_df.drop("Source", axis=1)
    result_df = result_df.rename(columns={"p-unc": "p-val"})
    result_df["reject"], result_df["p-val-adj"] = pg.multicomp(
        result_df["p-val"], method="fdr_bh"
    )
    return result_df


def _post_hoc(prepared_df: pd.DataFrame, trait_names: list[str]) -> pd.DataFrame:
    """Perform post-hoc test for the trait data."""
    post_hoc_results: list[pd.DataFrame] = []
    for trait_name in trait_names:
        post_hoc_result = pg.pairwise_gameshowell(
            data=prepared_df, dv=trait_name, between="group"
        )
        post_hoc_result.index = [trait_name] * len(post_hoc_result.index)
        post_hoc_results.append(post_hoc_result)
    return pd.concat(post_hoc_results)



"""This module contains the set of GlyTrait's exceptions."""


class GlyTraitError(Exception):
    """Base class for exceptions in this module."""


class GlycanParseError(GlyTraitError):
    """Raised when a glycan cannot be parsed."""


class StructureParseError(GlycanParseError):
    """Raised when a structure cannot be parsed."""


class CompositionParseError(GlycanParseError):
    """Raised when a composition cannot be parsed."""


class DataInputError(GlyTraitError):
    """The input file format error."""


class FileTypeError(GlyTraitError):
    """Raised when a file type is not supported."""


class NotEnoughGroupsError(GlyTraitError):
    """Raised when there are not enough groups."""


class SiaLinkageError(GlyTraitError):
    """Raised if a sialic acid linkage is not specified but used."""



"""Functions for loading data for GlyTrait.

This module provides functions for loading data for GlyTrait.

Classes:
    GlyTraitInputData: Encapsulates all the input data for GlyTrait.

Functions:
    load_input_data: Load all the input data for GlyTrait, including
        abundance table, glycans, and groups.
        Returns a `GlyTraitInputData` object.
"""

from __future__ import annotations

from collections.abc import Iterable, Callable
from typing import Optional, Literal, Protocol, cast, ClassVar

import pandas as pd
from attrs import define, field
from numpy import dtype

from glytrait.data_type import (
    AbundanceTable,
    GroupSeries,
)
from glytrait.exception import DataInputError
from glytrait.glycan import (
    parse_structures,
    parse_compositions,
    StructureDict,
    CompositionDict,
    GlycanDict,
)

__all__ = [
    "GlyTraitInputData",
    "load_input_data_from_csv",
]


# ===== Loaders =====
class AbundanceLoaderProto(Protocol):
    def load(self) -> AbundanceTable: ...


class GlycanLoaderProto(Protocol):
    def load(self) -> StructureDict | CompositionDict: ...


class GroupsLoaderProto(Protocol):
    def load(self) -> GroupSeries: ...


@define
class DFValidator:
    """Validator for pandas DataFrame.

    Attributes:
        must_have: List of column names that must be in the DataFrame.
        unique: List of column names that must be unique in the DataFrame.
        types: Dictionary of column names and their expected types.
        default_type: Default type for columns not in `types`.
    """

    must_have: list[str] = field(kw_only=True, factory=list)
    unique: list[str] = field(kw_only=True, factory=list)
    types: dict[str, str] = field(kw_only=True, factory=dict)
    default_type: dtype | str = field(kw_only=True, default=None)

    def __call__(self, df: pd.DataFrame) -> None:
        """Validate the DataFrame.

        Raises:
            DataInputError: If one of the following conditions is met:
            - The DataFrame does not have all the columns specified in `must_have`.
        """
        self._test_must_have_columns(df)
        self._test_unique_columns(df)
        self._test_type_check(df)
        self._test_default_type(df)

    def _test_must_have_columns(self, df: pd.DataFrame):
        if missing := {col for col in self.must_have if col not in df.columns}:
            msg = f"The following columns are missing: {', '.join(missing)}."
            raise DataInputError(msg)

    def _test_unique_columns(self, df: pd.DataFrame):
        if non_unique := [
            col for col in self.unique if col in df and df[col].duplicated().any()
        ]:
            msg = f"The following columns are not unique: {', '.join(non_unique)}."
            raise DataInputError(msg)

    def _test_type_check(self, df: pd.DataFrame):
        if wrong_type_cols := [
            col
            for col, dtype in self.types.items()
            if col in df and df[col].dtype != dtype
        ]:
            expected = {col: self.types[col] for col in wrong_type_cols}
            got = {col: df[col].dtype for col in wrong_type_cols}
            msg = (
                f"The following columns have incorrect types: {', '.join(wrong_type_cols)}. "
                f"Expected types: {expected}, got: {got}."
            )
            raise DataInputError(msg)

    def _test_default_type(self, df: pd.DataFrame):
        if self.default_type is None:
            return
        cols_to_check = set(df.columns) - set(self.types)
        if wrong_type_cols := [
            col for col in cols_to_check if df[col].dtype != self.default_type
        ]:
            got = {col: df[col].dtype for col in wrong_type_cols}
            msg = (
                f"The following columns have incorrect types: {', '.join(wrong_type_cols)}. "
                f"Expected types: {self.default_type}, got: {got}."
            )
            raise DataInputError(msg)


@define
class CSVLoader:
    """Base class for data loaders from csv files."""

    filepath: str
    validator: DFValidator = field(kw_only=True, default=lambda df: None)

    # Dependency injection
    reader: Callable[[str], pd.DataFrame] = field(kw_only=True, default=pd.read_csv)

    def load(self):
        """Load the data from the file."""
        raise NotImplementedError

    def load_df(self) -> pd.DataFrame:
        """Returns the DataFrame loaded from the file.

        Raises:
            FileNotFoundError: If the file does not exist.
            DataInputError: If one of the following conditions is met:
                (1) the file is empty; (2) the file could not be parsed.
        """
        df = self.read_csv()
        self.validator(df)
        return df

    def read_csv(self) -> pd.DataFrame:
        """A thin wrapper for `pd.read_csv`, with custum exceptions."""
        try:
            return self.reader(self.filepath)
        # FileNotFoundError is not caught here
        except pd.errors.EmptyDataError as e:
            raise DataInputError("Empty CSV file.") from e
        except pd.errors.ParserError as e:
            raise DataInputError("This CSV file could not be parsed.") from e


@define
class AbundanceCSVLoader(CSVLoader):
    """Loader for abundance table from a csv file."""

    validator: DFValidator = field(
        kw_only=True,
        default=DFValidator(
            must_have=["Sample"],
            unique=["Sample"],
            types={"Sample": "object"},
            default_type=dtype("float64"),
        ),
    )

    def load(self) -> AbundanceTable:
        """Returns the abundance table loaded from the file.

        Raises:
            FileNotFoundError: If the file does not exist.
            DataInputError: If the file format is incorrect.
                This includes: (1) the file is empty; (2) the file could not be parsed;
                (3) the "Sample" column is not found.
        """
        df = self.load_df()
        return AbundanceTable(df.set_index("Sample"))


@define
class GroupsCSVLoader(CSVLoader):
    """Loader for groups from a csv file."""

    validator: DFValidator = field(
        kw_only=True,
        default=DFValidator(
            must_have=["Group", "Sample"],
            unique=["Sample"],
            types={"Sample": "object"},
        ),
    )

    def load(self) -> GroupSeries:
        """Returns the groups loaded from the file.

        Raises:
            FileNotFoundError: If the file does not exist.
            DataInputError: If the file format is incorrect.
                This includes: (1) the file is empty; (2) the file could not be parsed;
                (3) the "Sample" column is not found; (4) the "Group" column is not found.
        """
        df = self.load_df()
        return GroupSeries(df.set_index("Sample")["Group"])


GlycanParserType = Callable[
    [Iterable[tuple[str, str]]], StructureDict | CompositionDict
]


@define
class GlycanCSVLoader(CSVLoader):
    """Loader for structures or compositions from a csv file."""

    mode: Literal["structure", "composition"] = field(kw_only=True)
    validator: DFValidator = field(kw_only=True, default=None)
    parser: GlycanParserType = field(kw_only=True, default=None)

    def __attrs_post_init__(self):
        if self.parser is None:
            self.parser = self._glycan_parser_factory(self.mode)
        if self.validator is None:
            self.validator = self._validator_factory(self.mode)

    @staticmethod
    def _glycan_parser_factory(
        mode: Literal["structure", "composition"]
    ) -> GlycanParserType:
        parser = parse_structures if mode == "structure" else parse_compositions
        return cast(GlycanParserType, parser)

    @staticmethod
    def _validator_factory(mode: Literal["structure", "composition"]) -> DFValidator:
        glycan_col = "Structure" if mode == "structure" else "Composition"
        return DFValidator(
            must_have=["GlycanID", glycan_col],
            unique=["GlycanID", glycan_col],
            types={"GlycanID": "object", glycan_col: "object"},
        )

    def load(self) -> GlycanDict:
        """Returns the glycans loaded from the file.

        Raises:
            FileNotFoundError: If the file does not exist.
            FileFormatError: If the file format is incorrect.
                This includes: (1) the file is empty; (2) the file could not be parsed;
                (3) the "GlycanID" column is not found; (4) the "Structure" or
                "Composition" column is not found.
            StructureParseError: If any structure cannot be parsed,
                when `mode` is "structure".
            CompositionParseError: If any composition cannot be parsed,
                when `mode` is "composition".
        """
        df = self.load_df()
        ids = df["GlycanID"].to_list()
        glycan_col = self.mode.capitalize()
        try:
            strings = df[glycan_col].to_list()
        except KeyError as e:
            raise DataInputError(f"The '{glycan_col}' column is not found.") from e
        return self.parser(zip(ids, strings))


# ===== Input data =====
@define(kw_only=True)
class GlyTraitInputData:
    """GlyTrait input data.

    Attributes:
        abundance_table: Abundance table as a pandas DataFrame.
        glycans: Glycans, either a dict of `Structure` objects or
            a dict of `Composition` objects.
        groups: Sample groups as a pandas Series.

    Notes:
        The glycan dict should have the same keys as the abundance table.
        The samples in the abundance table should have the same names as the
        samples in the groups.

    Raises:
        FileFormatError: If the abundance table has different samples as the groups,
            or if the glycan dict has different glycans as the abundance table.
    """

    abundance_table: AbundanceTable
    glycans: StructureDict | CompositionDict
    groups: Optional[GroupSeries] = None


# ===== Input data validators =====
ValidatorType = Callable[[GlyTraitInputData], None]


@define
class InputDataValidator:
    """Validator for the input data.

    This class only validates the interaction between the input data.
    The format of each input data is validated by the loaders.
    """

    validators: ClassVar[list[ValidatorType]] = []

    def __call__(self, input_data: GlyTraitInputData) -> None:
        for validator in self.validators:
            validator(input_data)

    @classmethod
    def add_validator(cls, validator: ValidatorType) -> ValidatorType:
        """A decorator to add a validator to the list of validators."""
        cls.validators.append(validator)
        return validator


@InputDataValidator.add_validator
def same_samples_in_abundance_and_groups(input_data: GlyTraitInputData) -> None:
    """Check if the abundance table and the groups have the same samples.

    Raises:
        DataInputError: If the samples in the abundance table and the groups are different.
    """
    if input_data.groups is None:
        return
    abund_samples = set(input_data.abundance_table.index)
    groups_samples = set(input_data.groups.index)
    if abund_samples != groups_samples:
        samples_in_abund_not_in_groups = abund_samples - groups_samples
        samples_in_groups_not_in_abund = groups_samples - abund_samples
        msg = ""
        if samples_in_abund_not_in_groups:
            msg += (
                f"The following samples are in the abundance table but not in the groups: "
                f"{', '.join(samples_in_abund_not_in_groups)}. "
            )
        if samples_in_groups_not_in_abund:
            msg += (
                f"The following samples are in the groups but not in the abundance table: "
                f"{', '.join(samples_in_groups_not_in_abund)}."
            )
        raise DataInputError(msg)


@InputDataValidator.add_validator
def all_glycans_have_structures_or_compositions(input_data: GlyTraitInputData) -> None:
    """Check if all glycans in the abundance table have structures or compositions.

    Glycans in the structure or composition dict that are not in the abundance table
    are not checked.

    Raises:
        DataInputError: If any glycan in the abundance table does not
            have a structure or composition.
    """
    abund_glycans = set(input_data.abundance_table.columns)
    glycans = set(input_data.glycans.keys())
    if diff := abund_glycans - glycans:
        msg = (
            f"The following glycans in the abundance table do not have structures or "
            f"compositions: {', '.join(diff)}."
        )
        raise DataInputError(msg)


# ===== The low-level API =====
def load_input_data(
    *,
    abundance_loader: AbundanceLoaderProto,
    glycan_loader: GlycanLoaderProto,
    group_loader: Optional[GroupsLoaderProto] = None,
    validator: ValidatorType = InputDataValidator(),
) -> GlyTraitInputData:
    """Load all the input data for GlyTrait.

    Args:
        abundance_loader: Loader for the abundance table.
        glycan_loader: Loader for the glycans.
        group_loader: Loader for the groups. Optional.
        mode: Either "structure" or "composition".
        validator: Validator for the input data.

    Returns:
        GlyTraitInputData: Input data for GlyTrait.
    """
    abundance_table = abundance_loader.load()
    glycans = glycan_loader.load()
    groups = group_loader.load() if group_loader else None

    input_data = GlyTraitInputData(
        abundance_table=abundance_table,
        glycans=glycans,
        groups=groups,
    )

    validator(input_data)
    return input_data


# ===== The high-level API =====
def load_input_data_from_csv(
    *,
    abundance_file: str,
    glycan_file: str,
    group_file: Optional[str] = None,
    mode: Literal["structure", "composition"],
) -> GlyTraitInputData:
    """Load all the input data for GlyTrait from csv files.

    Args:
        abundance_file: Path to the abundance table csv file.
        glycan_file: Path to the glycans csv file.
        group_file: Path to the groups csv file. Optional.
        mode: Either "structure" or "composition".

    Returns:
        GlyTraitInputData: Input data for GlyTrait.
    """
    abundance_loader = AbundanceCSVLoader(filepath=abundance_file)
    glycan_loader = GlycanCSVLoader(filepath=glycan_file, mode=mode)
    group_loader = GroupsCSVLoader(filepath=group_file) if group_file else None
    return load_input_data(
        abundance_loader=abundance_loader,
        glycan_loader=glycan_loader,
        group_loader=group_loader,
    )



"""Preprocess the abundance table.

The only high-level function that client code will use is `preprocess`.
"""

from typing import Literal, Protocol

from attrs import define

from glytrait.load_data import GlyTraitInputData
from glytrait.data_type import AbundanceTable

__all__ = ["preprocess"]


# This is the high-level function that client code will use.
def preprocess(
    data: GlyTraitInputData,
    filter_max_na: float,
    impute_method: Literal["zero", "min", "lod", "mean", "median"],
) -> None:
    """Preprocess the abundance table.

    Notes:
        This function will modify the input data in place.

    Args:
        data (GlyTraitInputData): The input data.
        filter_max_na (float): The maximum proportion of missing values allowed for a glycan.
        impute_method (Literal["zero", "min", "lod", "mean", "median"]): The imputation method.
    """
    steps = [
        FilterGlycans(max_na=filter_max_na),
        Impute(method=impute_method),
        Normalize(),
    ]
    pipeline = ProcessingPipeline(steps=steps)
    pipeline(data)


class ProcessingStep(Protocol):
    """The protocol for processing steps."""

    def __call__(self, data: GlyTraitInputData) -> None: ...


@define
class ProcessingPipeline:
    """The pipeline for processing the abundance table."""

    _steps: list[ProcessingStep]

    def __call__(self, data: GlyTraitInputData) -> None:
        for step in self._steps:
            step(data)


@define
class FilterGlycans(ProcessingStep):
    """Filter glycans with too many missing values.

    Args:
        max_na (float): The maximum proportion of missing values allowed for a glycan.
    """

    max_na: float

    def __call__(self, data: GlyTraitInputData) -> None:
        glycans_to_keep = _filter_glycans(data.abundance_table, self.max_na)
        data.abundance_table = AbundanceTable(
            data.abundance_table.filter(items=glycans_to_keep)
        )
        for glycan in set(data.glycans).difference(glycans_to_keep):
            del data.glycans[glycan]


def _filter_glycans(abundance_df: AbundanceTable, max_na: float) -> list[str]:
    to_keep_mask = abundance_df.isna().mean() <= max_na
    glycans_to_keep = list(abundance_df.columns[to_keep_mask])
    return glycans_to_keep


@define
class Impute(ProcessingStep):
    """Impute the missing values of the abundance table.

    The following imputation methods supported:
        - "zero": Replace the missing values with 0.
        - "min": Replace the missing values with the minimum value of the corresponding glycan.
        - "lod": Replace the missing values with 1/5 of the minimum value of the corresponding
            glycan.
        - "mean": Replace the missing values with the mean value of the corresponding glycan.
        - "median": Replace the missing values with the median value of the corresponding glycan.

    Args:
        method (str): The imputation method.
            Can be "zero", "min", "lod", "mean", "median".
    """

    method: Literal["zero", "min", "lod", "mean", "median"]

    def __call__(self, data: GlyTraitInputData) -> None:
        data.abundance_table = _impute(data.abundance_table, self.method)


def _impute(
    abundance_df: AbundanceTable,
    method: Literal["zero", "min", "lod", "mean", "median"],
) -> AbundanceTable:
    if method == "zero":
        imputed_df = abundance_df.fillna(0)
    elif method == "min":
        imputed_df = abundance_df.fillna(abundance_df.min())
    elif method == "lod":
        imputed_df = abundance_df.fillna(abundance_df.min() / 5)
    elif method == "mean":
        imputed_df = abundance_df.fillna(abundance_df.mean())
    elif method == "median":
        imputed_df = abundance_df.fillna(abundance_df.median())
    else:
        raise ValueError("The imputation method is not supported.")
    return AbundanceTable(imputed_df)


@define
class Normalize(ProcessingStep):
    """Normalize the abundance table by dividing the sum of each sample."""

    def __call__(self, data: GlyTraitInputData) -> None:
        data.abundance_table = _normalization(data.abundance_table)


def _normalization(abundance_df: AbundanceTable) -> AbundanceTable:
    """Normalize the abundance table by dividing the sum of each sample.

    Args:
        abundance_df (AbundanceTable): The abundance table,
            with samples as index and Compositions as columns.

    Returns:
        normalized_df (AbundanceTable): The normalized abundance table.
    """
    row_sums = abundance_df.sum(axis=1)
    normalized_df = abundance_df.div(row_sums, axis=0).round(6)
    return AbundanceTable(normalized_df)



"""This module implements classes about representing glycans.

Classes:
    Structure: A class representing a glycan structure.
    Composition: A class representing a glycan composition.

Functions:
    parse_structures: Parse a list of glycan structures.
    parse_compositions: Parse a list of glycan compositions.
    get_mono_str: Helper function for checking the identity of a `glypy.MonosaccarideResidue`.
"""

from __future__ import annotations

import re
from collections.abc import Generator, Mapping, Iterator, Callable
from typing import Literal, Iterable, Optional, Final, Union, TypeVar, NewType

from attrs import frozen, field
from glypy.io.glycoct import loads as glycoct_loads, GlycoCTError  # type: ignore
from glypy.structure.glycan import Glycan as GlypyGlycan  # type: ignore
from glypy.structure.glycan_composition import (  # type: ignore
    GlycanComposition,
    MonosaccharideResidue,
)

from glytrait.exception import (
    GlycanParseError,
    StructureParseError,
    CompositionParseError,
)

__all__ = [
    "StructureDict",
    "CompositionDict",
    "GlycanDict",
    "parse_structures",
    "parse_compositions",
    "Structure",
    "Composition",
    "get_mono_str",
]

StructureDict = NewType("StructureDict", dict[str, "Structure"])
"""A type hint for a dictionary of `Structure` instances.
The keys are the names of the structures, and the values are the structures.
"""

CompositionDict = NewType("CompositionDict", dict[str, "Composition"])
"""A type hint for a dictionary of `Composition` instances.
The keys are the names of the compositions, and the values are the compositions.
"""

GlycanDict = Union[StructureDict, CompositionDict]
"""A type hint for a dictionary of `Structure` or `Composition` instances.
The keys are the names of the structures or compositions, 
and the values are the structures or compositions.
"""


def parse_structures(__iter: Iterable[tuple[str, str]]) -> StructureDict:
    """Parse glycan structures from a list of structure strings.

    Args:
        An iterable of tuples of (name, structure_string).

    Returns:
        A dictionary of `Structure` instances, with the keys being the names of the structures.

    Raises:
        StructureParseError: When the string cannot be parsed.
            The names of the structures that cannot be parsed are included
            in the error message.
    """
    try:
        result: dict[str, Structure] = _load_glycans(__iter, Structure.from_string)
    except GlycanParseError as exc:
        raise StructureParseError(f"Could not parse structures for: {exc}.")
    else:
        return StructureDict(result)


def parse_compositions(__iter: Iterable[tuple[str, str]]) -> CompositionDict:
    """Parse glycan compositions from a list of composition strings.

    Args:
        An iterable of tuples of (name, composition_string).

    Returns:
        A dictionary of `Composition` instances, with the keys being the names of the compositions.

    Raises:
        CompositionParseError: When the string cannot be parsed.
            The names of the compositions that cannot be parsed are included
            in the error message.
    """
    try:
        result: dict[str, Composition] = _load_glycans(__iter, Composition.from_string)
    except GlycanParseError as exc:
        raise CompositionParseError(f"Could not parse compositions for: {exc}.")
    else:
        return CompositionDict(result)


Glycan = TypeVar("Glycan", bound=Union["Structure", "Composition"])
GlycanBuilder = Callable[[str, str], Glycan]


def _load_glycans(
    __iter: Iterable[tuple[str, str]], builder: GlycanBuilder
) -> dict[str, Glycan]:
    failed_names: list[str] = []
    glycans: dict[str, Glycan] = {}
    for name, string in __iter:
        try:
            glycan = builder(name, string)
        except GlycanParseError:
            failed_names.append(name)
        else:
            glycans[name] = glycan
    if failed_names:
        msg = ", ".join(f"'{name}'" for name in failed_names)
        raise StructureParseError(msg)
    return glycans


@frozen
class Structure:
    """The structure of a glycan.

    Attributes:
        name (str): The name of the glycan.
        composition (dict): The composition of the glycan.

    Methods:
        from_string: classmethod for building a `Structure` instance from a string.
        from_glycoct: classmethod for building a `Structure` instance from a glycoCT string.
        breadth_first_traversal: traverse the structure in a "bfs" manner.
        depth_first_traversal: traverse the structure in a "dfs" manner.
    """

    name: str = field()
    _glypy_glycan: GlypyGlycan = field(repr=False)
    _composition: dict[str, int] = field(init=False, repr=False, hash=False)

    def __attrs_post_init__(self):
        self._init_composition()

    def _init_composition(self):
        glypy_comp = GlycanComposition.from_glycan(self._glypy_glycan)
        comp = {str(k): v for k, v in glypy_comp.items()}
        object.__setattr__(self, "_composition", comp)

    @classmethod
    def from_string(
        cls, name: str, string: str, *, format: Literal["glycoct"] = "glycoct"
    ) -> Structure:
        """Build a glycan from a string representation.

        Args:
            name (str): The name of the glycan.
            string (str): The string representation of the glycan.
            format (Literal["glycoct"], optional): The format of the string.
                Defaults to "glycoct".

        Returns:
            Structure: The glycan.

        Raises:
            StructureParseError: When the string cannot be parsed.
        """
        if format == "glycoct":
            try:
                return cls(name, glycoct_loads(string))
            except GlycoCTError:
                raise StructureParseError(f"Could not parse string: {string}")
        else:
            raise StructureParseError(f"Unknown format: {format}")

    @classmethod
    def from_glycoct(cls, name: str, glycoct: str) -> Structure:
        """Build a glycan from a GlycoCT string.

        Args:
            name (str): The name of the glycan.
            glycoct (str): The GlycoCT string.

        Returns:
            NGlycan: The glycan.

        Raises:
            StructureParseError: When the string cannot be parsed.
        """
        return cls.from_string(name, glycoct, format="glycoct")

    def _traversal(
        self,
        method: Literal["bfs", "dfs"],
        *,
        skip: Optional[Iterable[str]] = None,
        only: Optional[Iterable[str]] = None,
    ) -> Generator[MonosaccharideResidue, None, None]:
        # set the traversal method
        if method == "bfs":
            traversal_func = self._glypy_glycan.breadth_first_traversal
        elif method == "dfs":
            traversal_func = self._glypy_glycan.depth_first_traversal
        else:
            raise ValueError(f"Unknown traversal method: {method}")

        # check the validation of `skip` and `only`
        if skip and only:
            raise ValueError("Cannot specify both `skip` and `only`.")

        # traverse the glycan
        if skip is None and only is None:
            yield from traversal_func()
        elif skip is not None:
            for node in traversal_func():
                if get_mono_str(node) not in skip:
                    yield node
        else:  # only is not None
            for node in traversal_func():
                if get_mono_str(node) in only:  # type: ignore
                    yield node

    def breadth_first_traversal(
        self,
        *,
        skip: Optional[Iterable[str]] = None,
        only: Optional[Iterable[str]] = None,
    ) -> Generator[MonosaccharideResidue, None, None]:
        """Traverse the structure in a "bfs" manner.

        Args:
            skip (Optional[Iterable[str]], optional): The monosaccharides to skip.
                Defaults to None.
            only (Optional[Iterable[str]], optional): The monosaccharides to traverse.
                Defaults to None.

        Yields:
            glypy.MonosaccharideResidue: The monosaccharide residues.
        """
        yield from self._traversal("bfs", skip=skip, only=only)

    def depth_first_traversal(
        self,
        *,
        skip: Optional[Iterable[str]] = None,
        only: Optional[Iterable[str]] = None,
    ) -> Generator[MonosaccharideResidue, None, None]:
        """Traverse the structure in a "dfs" manner.

        Args:
            skip (Optional[Iterable[str]], optional): The monosaccharides to skip.
                Defaults to None.
            only (Optional[Iterable[str]], optional): The monosaccharides to traverse.
                Defaults to None.

        Yields:
            glypy.MonosaccharideResidue: The monosaccharide residues.
        """
        yield from self._traversal("dfs", skip=skip, only=only)

    @property
    def composition(self) -> dict[str, int]:
        """The composition of the glycan."""
        return self._composition.copy()


VALID_MONOS: Final = ["H", "N", "F", "S", "L", "E"]


# The order here are used in the string representation of a composition.


@frozen
class Composition(Mapping[str, int]):
    """A glycan composition.

    Valid monosaccharides are: H, N, F, S, L, E.
    S must not be with L or E.
    Numbers of monosaccharides must be above 0.

    Attributes:
        name (str): The name of the glycan.

    Methods:
        from_string: classmethod to build a `Composition` instance from a string.

    Examples:
        >>> comp = Composition("G", {"H": 5, "N": 4, "F": 1})
        >>> comp
        Composition(name='G', comp={'H': 5, 'N': 4, 'F': 1})
        >>> str(comp)
        "H5N4F1"
        >>> comp["H"]
        5
        >>> comp["S"]
        0  # note that 'S' is not used when initialization
        >>> comp["P"]
        # raise KeyError
        >>> comp["H"] = 6
        # raise TypeError (unchangeable)
        >>> comp.get("P", default=0)
        0  # default value
        >>> len(comp)
        10  # 5 + 4 + 1
        >>> for mono, num in comp.items():
        ...     print(mono, num)
        H 5
        N 4
        F 1
        >>> Composition("G", {"H": 5, "N":4, "S": 1, "L": 1})
        # raise CompositionParseError (S must not be with L or E)
    """

    name: str = field()
    _comp: dict[str, int] = field(converter=dict, hash=False)

    def __attrs_post_init__(self):
        self._validate_comp()
        self._remove_zero()
        self._reorder()

    def _validate_comp(self) -> None:
        """Validate the composition."""
        self._validate_mono()
        self._validate_num()
        self._validate_sia()

    def _validate_mono(self) -> None:
        """Make sure that the monosaccharides are valid."""
        for k in self._comp:
            if k not in VALID_MONOS:
                raise CompositionParseError(f"Unknown monosaccharide: {k}.")

    def _validate_num(self) -> None:
        """Make sure that the numbers are valid."""
        for k, v in self._comp.items():
            if v < 0:
                raise CompositionParseError(f"Monosacharride must be above 0: {k}={v}.")

    def _validate_sia(self) -> None:
        """Make sure that S is not with L or E."""
        has_S = self._comp.get("S", 0) > 0
        has_L = self._comp.get("L", 0) > 0
        has_E = self._comp.get("E", 0) > 0
        if has_S and (has_L or has_E):
            raise CompositionParseError("S must not be with L or E.")

    def _remove_zero(self) -> None:
        to_delete: set[str] = set()
        for k, v in self._comp.items():
            if v == 0:
                to_delete.add(k)
        for k in to_delete:
            del self._comp[k]

    def _reorder(self) -> None:
        new_comp: dict[str, int] = {}
        for mono in VALID_MONOS:
            if mono in self._comp:
                new_comp[mono] = self._comp[mono]
        object.__setattr__(self, "_comp", new_comp)

    @classmethod
    def from_string(cls, name: str, string: str) -> Composition:
        """Create a composition from a string.

        Args:
            name (str): The name of the glycan.
            string (str): The string representation of the composition.

        Returns:
            Composition: The composition.

        Raises:
            CompositionParseError: When the string cannot be parsed.
        """
        if string == "":
            raise CompositionParseError("Empty string.")
        pattern = r"^([A-Z]\d+)*$"
        if not re.fullmatch(pattern, string):
            raise CompositionParseError(f"Invalid composition: {string}.")
        mono_comp: dict[str, int] = {}
        pattern = r"([A-Z])(\d+)"
        for m in re.finditer(pattern, string):
            mono_comp[m.group(1)] = int(m.group(2))
        return cls(name, mono_comp)  # type: ignore

    def __getitem__(self, __key: str) -> int:
        try:
            return self._comp[__key]
        except KeyError:
            if __key in VALID_MONOS:
                return 0
            else:
                raise

    def __len__(self) -> int:
        return sum(self._comp.values())

    def __iter__(self) -> Iterator[str]:
        return iter(self._comp)

    def __str__(self) -> str:
        mono_strings: list[str] = []
        for mono, num in self._comp.items():
            mono_strings.append(f"{mono}{num}")
        return "".join(mono_strings)


def get_mono_str(mono: MonosaccharideResidue) -> str:
    """Get the string representation of a monosaccharide residue.

    This is a helper function for checking the identity of a `glypy.MonosaccarideResidue`
    instance.

    Args:
        mono (glypy.MonosaccharideResidue): The monosaccharide residue.
    """
    return MonosaccharideResidue.from_monosaccharide(mono).name()



"""Post-filtering the derived traits table.

The only function in this module is `post_filter`.
"""

from typing import Iterable, Literal

import numpy as np

from glytrait.formula import TraitFormula
from glytrait.data_type import DerivedTraitTable

__all__ = ["post_filter"]


def post_filter(
    formulas: Iterable[TraitFormula],
    trait_df: DerivedTraitTable,
    threshold: float,
    method: Literal["pearson", "spearman"],
) -> DerivedTraitTable:
    """Post-filter the derived traits table.

    The post-filtering consists of two steps:
    1. Rule out the invalid traits.
    2. Reduce the colinearity of the traits.

    Args:
        formulas (Iterable[TraitFormula]): The formulas to be filtered.
        trait_df (DerivedTraitTable): The derived traits table.
        threshold (float): The threshold of the correlation coefficient.
            If set to -1, the colinearity filtering will be skipped.
        method (Literal["pearson", "spearman"]): The method to calculate the correlation.

    Returns:
        DerivedTraitTable: The filtered derived traits table.
    """
    trait_df = filter_invalid(trait_df)
    if threshold != -1:
        trait_df = filter_colinearity(formulas, trait_df, threshold, method)
    return trait_df


def filter_invalid(trait_df: DerivedTraitTable) -> DerivedTraitTable:
    """Rule out the invalid traits.

    A trait is invalid if it:
    1. Has the same value for all samples.
    2. Is NaN for all samples.

    Args:
        trait_df (DerivedTraitTable): The derived traits table.

    Returns:
        DerivedTraitTable: The filtered trait values.
    """
    trait_df = _filter_all_same(trait_df)
    trait_df = _filter_all_nan(trait_df)
    return trait_df


def _filter_all_same(trait_df: DerivedTraitTable) -> DerivedTraitTable:
    """Rule out the traits that have the same value for all samples."""
    return DerivedTraitTable(trait_df.loc[:, trait_df.nunique() != 1])


def _filter_all_nan(trait_df: DerivedTraitTable) -> DerivedTraitTable:
    """Rule out the traits that are NaN for all samples."""
    return DerivedTraitTable(trait_df.loc[:, trait_df.notna().any()])


def filter_colinearity(
    formulas: Iterable[TraitFormula],
    trait_df: DerivedTraitTable,
    threshold: float,
    method: Literal["pearson", "spearman"],
) -> DerivedTraitTable:
    """Filter the colinearity of the formulas.

    Args:
        formulas (Iterable[TraitFormula]): The formulas to be filtered.
        trait_df (DerivedTraitTable): The derived traits table, after post-filtering.
        threshold (float): The threshold of the correlation coefficient.
        method (Literal["pearson", "spearman"]): The method to calculate the correlation.

    Returns:
        DerivedTraitTable: The filtered derived traits table.
    """
    trait_names = list(trait_df.columns)

    # First, build the parent-child relationship matrix.
    # This matrix consists of 0 and 1.
    # If the i-th row and j-th column is 1, then the i-th trait is the child of the j-th trait.
    rela_matrix = _relationship_matrix(trait_names, formulas)

    # Second, build the correlation matrix.
    # This matrix also consists of 0 and 1.
    # 1 means the two traits are highly correlated (r > the threshold).
    corr_matrix = _correlation_matrix(trait_df, threshold, method)

    # Third, multiply the two matrices.
    # If the i-th row and j-th column is 1,
    # then the i-th trait is the child of the j-th trait,
    # and the two traits are highly correlated.
    # Then the i-th trait should be removed.
    # That is to say, if the row sum of the i-th row is larger than 0,
    # then the i-th trait should be removed.
    remove_matrix = rela_matrix * corr_matrix
    to_keep = np.sum(remove_matrix, axis=1) == 0

    # Finally, filter the formulas and the derived traits table.
    filtered_trait_table = trait_df.loc[:, to_keep]
    return DerivedTraitTable(filtered_trait_table)


def _relationship_matrix(trait_names: Iterable[str], formulas: Iterable[TraitFormula]):
    """Build the parent-child relationship matrix.

    This matrix consists of 0 and 1.
    If the i-th row and j-th column is 1, then the i-th trait is the child of the j-th trait.

    Args:
        trait_names (Iterable[str]): The names of the traits.
        formulas (Iterable[TraitFormula]): The formulas to be filtered.

    Returns:
        np.ndarray: The relationship matrix.
    """
    trait_names = list(trait_names)
    formula_dict = {
        formula.name: formula for formula in formulas if formula.name in trait_names
    }
    matrix = np.zeros((len(trait_names), len(trait_names)), dtype=int)
    for i, trait_1 in enumerate(trait_names):
        for j, trait_2 in enumerate(trait_names):
            formula_1 = formula_dict[trait_1]
            formula_2 = formula_dict[trait_2]
            if _is_child_of(formula_1, formula_2):
                matrix[i, j] = 1
    return matrix


def _correlation_matrix(
    trait_table: DerivedTraitTable,
    threshold: float,
    method: Literal["pearson", "spearman"],
):
    """Build the correlation matrix.

    This matrix also consists of 0 and 1.
    1 means the two traits are highly correlated (r > the threshold).

    Args:
        trait_table (DerivedTraitTable): The derived traits table, after post-filtering.
        threshold (float): The threshold of the correlation coefficient.
        method (Literal["pearson", "spearman"]): The method to calculate the correlation.

    Returns:
        np.ndarray: The correlation matrix.
    """
    corr_matrix = trait_table.corr(method=method).values >= threshold
    corr_matrix = corr_matrix.astype(int)
    return corr_matrix


def _is_child_of(trait1: TraitFormula, trait2: TraitFormula) -> bool:
    """Whether trait1 is a child of the trait2.

    A formula is a child of another formula
    if both numerator and denominator of this formula have the same additional
    term as the other formula.

    For example,
    Formula 1:
      - Numerators: A, B, C
      - Denominators: A, B
    Formula 2:
      - Numerators: A, B, C, D
      - Denominators: A, B, D
    Then Formula 2 is a child of Formula 1.

    """
    num1 = set(t.expr for t in trait1.numerators)
    den1 = set(t.expr for t in trait1.denominators)
    try:
        num2 = set(t.expr for t in trait2.numerators)
        den2 = set(t.expr for t in trait2.denominators)
    except AttributeError:
        raise TypeError("The other formula is not a TraitFormula instance.")

    condition = (
        not (num2 - num1)
        and not (den2 - den1)
        and num1 - num2 == den1 - den2
        and len(num1 - num2) == 1
    )

    return condition



from __future__ import annotations

from typing import NewType

import pandas as pd

AbundanceTable = NewType("AbundanceTable", pd.DataFrame)
"""Abundance table type.
The index are samples and the columns are glycans.
The abundance table could only be returned by `load_abundance_table` function.
"""

GroupSeries = NewType("GroupSeries", pd.Series)
"""Group series type.
The index are the sample names and the values are groups.
"""

MetaPropertyTable = NewType("MetaPropertyTable", pd.DataFrame)
"""The type of the meta-property table. 
Only returned by `build_meta_property_table`.
A pandas DataFrame with `glycan_ids` as the index, 
and the meta-property names as the columns.
"""

DerivedTraitTable = NewType("DerivedTraitTable", pd.DataFrame)
"""The type of the derived trait table.
Only returned by `calcu_derived_trait`.
A pandas DataFrame with samples as the index,
and the derived trait names as the columns.
"""



from typing import cast, Literal, Optional

import pandas as pd
from attrs import define, field

from glytrait.data_export import export_all
from glytrait.data_type import MetaPropertyTable, DerivedTraitTable, GroupSeries
from glytrait.formula import TraitFormula, load_formulas, load_default_formulas
from glytrait.load_data import load_input_data_from_csv, GlyTraitInputData
from glytrait.meta_property import build_meta_property_table
from glytrait.post_filtering import post_filter
from glytrait.preprocessing import preprocess
from glytrait.trait import calcu_derived_trait
from glytrait.stat import t_test, anova


class GlyTrait:
    """GlyTrait API.

    Args:
        mode (str): "structure" or "composition". Default: "structure".
        filter_max_na (float): The maximum ratio of missing values in a sample.
            If the ratio of missing values in a sample is greater than this value,
            the sample will be removed.
            Setting to 1.0 means no filtering.
            Setting to 0.0 means only keeping glycans with no missing values.
            Default: 1.0.
        impute_method (str): The method to impute missing values.
            "zero": fill with 0.
            "min": fill with the minimum value of the column.
            "lod": fill with the limit of detection of the column.
            "mean": fill with the mean value of the column.
            "median": fill with the median value of the column.
            Default: "zero".
        post_filtering (bool): Whether to perform post filtering.
            If True, the invalid traits and the highly correlated traits will be removed.
            Default: True.
        correlation_threshold (float): The correlation threshold for post filtering.
            If the correlation between two traits is greater than this value,
            one of them will be removed.
            Setting to -1.0 means no correlation filtering.
            Default: 1.0.
        sia_linkage (bool): Whether to consider the linkage of sialic acid.
            If True, the linkage of sialic acid will be considered in the calculation of
            meta properties. Default: False.
        custom_formula_file (str): Path to the custom formula file.

    Examples:
        >>> from glytrait import GlyTrait
        >>> glytrait = GlyTrait()
        >>> glytrait.run(
        ...     output_dir="output",
        ...     abundance_file="glycan_abundance.csv",
        ...     glycan_file="glycan_structure.csv",
        ...     group_file="group.csv",
        ... )
    """

    def __init__(self, **kwargs):
        self._config = _Config(**kwargs)
        self._formulas = self._init_formulas()

    def _init_formulas(self) -> list[TraitFormula]:
        if self._config.custom_formula_file is not None:
            return load_formulas(
                self._config.custom_formula_file, self._config.sia_linkage
            )
        else:
            return load_default_formulas(self._config.mode, self._config.sia_linkage)

    def run(
        self,
        output_dir: str,
        abundance_file: str,
        glycan_file: str,
        group_file: Optional[str] = None,
    ):
        """Run GlyTrait.

        Three files are required: abundance_file, glycan_file, and group_file.

        The abundance file is a csv file with the first column as the sample names, and the
        remaining columns as the abundance values for different glycans.
        The first column should be named "Sample".

        Two choices for the glycan file:
        1. A csv file with the first column "GlycanID" as the glycan names,
        and the second column "Structure" or "Composition" as the glycan
        structure or composition.
        2. A directory with each file named as the glycan names, and the file
        content as the glycan structure.
        The second choice is only available when mode is "structure".
        At this moment, only "glycoct" format is supported.
        So all files in this directory should have the extension ".glycoct".

        The group file is a csv file with the first column "Sample" as the sample names,
        and the second column "Group" as the group names.

        Args:
            output_dir (str): Path to the output directory.
            abundance_file (str): Path to the abundance file.
            glycan_file (str): Path to the glycan file.
            group_file (str): Path to the group file. Optional.
        """
        input_data = self._load_input_data(abundance_file, glycan_file, group_file)
        self._preprocess(input_data)
        meta_property_table = self._calcu_meta_property(input_data)
        derived_trait_table = self._calcu_derived_trait(input_data, meta_property_table)
        if self._config.post_filtering:
            filtered_derived_trait_table = self._post_filtering(derived_trait_table)
        else:
            filtered_derived_trait_table = None
        if group_file is not None and self._config.post_filtering:
            diff_result = self._diff_analysis(filtered_derived_trait_table, input_data)
        else:
            diff_result = {}
        self._export_data(
            output_dir,
            input_data,
            meta_property_table,
            derived_trait_table,
            filtered_derived_trait_table,
            diff_result,
        )

    def _load_input_data(
        self, abundance_file: str, glycan_file: str, group_file: Optional[str] = None
    ) -> GlyTraitInputData:
        return load_input_data_from_csv(
            abundance_file=abundance_file,
            glycan_file=glycan_file,
            group_file=group_file,
            mode=self._config.mode,
        )

    def _preprocess(self, input_data: GlyTraitInputData) -> None:
        preprocess(input_data, self._config.filter_max_na, self._config.impute_method)

    def _calcu_meta_property(self, input_data: GlyTraitInputData) -> MetaPropertyTable:
        return build_meta_property_table(
            input_data.glycans, self._config.mode, self._config.sia_linkage
        )

    def _calcu_derived_trait(
        self, input_data: GlyTraitInputData, meta_property_table: MetaPropertyTable
    ) -> DerivedTraitTable:
        return calcu_derived_trait(
            input_data.abundance_table, meta_property_table, self._formulas
        )

    def _post_filtering(
        self, derived_trait_table: DerivedTraitTable
    ) -> DerivedTraitTable:
        return post_filter(
            formulas=self._formulas,
            trait_df=derived_trait_table,
            threshold=self._config.correlation_threshold,
            method="pearson",
        )

    def _diff_analysis(
        self,
        derived_trait_table: DerivedTraitTable | None,
        input_data: GlyTraitInputData,
    ) -> dict[str, pd.DataFrame]:
        groups = cast(GroupSeries, input_data.groups)
        trait_table = cast(DerivedTraitTable, derived_trait_table)
        if groups.unique().size == 2:
            return {"t_test.csv": t_test(trait_table, groups)}
        else:  # groups size > 2
            anova_df, post_hoc_df = anova(trait_table, groups)
            return {"anova.csv": anova_df, "post_hoc.csv": post_hoc_df}

    def _export_data(
        self,
        output_dir: str,
        input_data: GlyTraitInputData,
        meta_property_table: MetaPropertyTable,
        derived_trait_table: DerivedTraitTable,
        filtered_derived_trait_table: Optional[DerivedTraitTable] = None,
        diff_results: Optional[dict[str, pd.DataFrame]] = None,
    ) -> None:
        data_to_export = [
            ("meta_properties.csv", meta_property_table),
            ("derived_traits.csv", derived_trait_table),
            ("glycan_abundance_processed.csv", input_data.abundance_table),
        ]
        if filtered_derived_trait_table is not None:
            data_to_export.append(
                ("derived_traits_filtered.csv", filtered_derived_trait_table)
            )
        if diff_results is not None:
            data_to_export.extend(diff_results.items())
        export_all(data_to_export, output_dir)


@define
class _Config:
    """GlyTrait configuration.

    This class encapsulates the configuration validation logic.
    """

    mode: Literal["structure", "composition"] = field(default="structure")
    filter_max_na: float = field(default=1.0)
    impute_method: Literal["zero", "min", "lod", "mean", "median"] = field(
        default="zero"
    )
    post_filtering: bool = field(default=True)
    correlation_threshold: float = field(default=1.0)
    sia_linkage: bool = field(default=False)
    custom_formula_file: Optional[str] = field(default=None)

    @mode.validator
    def _validate_mode(self, attribute, value):  # type: ignore
        if not isinstance(value, str):
            raise ValueError("mode must be a string.")
        if value not in {"structure", "composition"}:
            raise ValueError("mode must be one of: structure, composition.")

    @filter_max_na.validator
    def _validate_filter_max_na(self, attribute, value):  # type: ignore
        if not isinstance(value, (float, int)):
            raise ValueError("filter_max_na must be a float.")
        if not 0 <= value <= 1:
            raise ValueError("filter_max_na must be between 0 and 1.")

    @impute_method.validator
    def _validate_impute_method(self, attribute, value):  # type: ignore
        if not isinstance(value, str):
            raise ValueError("impute_method must be a string.")
        if value not in {"zero", "min", "lod", "mean", "median"}:
            raise ValueError(
                "impute_method must be one of: zero, min, lod, mean, median."
            )

    @post_filtering.validator
    def _validate_post_filtering(self, attribute, value):  # type: ignore
        if not isinstance(value, bool):
            raise ValueError("post_filtering must be a boolean.")

    @correlation_threshold.validator
    def _validate_correlation_threshold(self, attribute, value):  # type: ignore
        if not isinstance(value, (float, int)):
            raise ValueError("correlation_threshold must be a float.")
        if not (0 <= value <= 1 or value == -1):
            raise ValueError("correlation_threshold must be between 0 and 1, or -1.")

    @sia_linkage.validator
    def _validate_sia_linkage(self, attribute, value):  # type: ignore
        if not isinstance(value, bool):
            raise ValueError("sia_linkage must be a boolean.")

    @custom_formula_file.validator
    def _validate_custom_formula_file(self, attribute, value):  # type: ignore
        if value is not None:
            if not isinstance(value, str):
                raise ValueError("custom_formula_file must be a string.")



"""Export data to files.

This module contains a function `export_all` that exports all data to files.
The data is provided as an Iterable of tuples (filename, data object).
This module will handle different data types by using the appropriate exporter.

Examples:
    >>> from glytrait.data_export import export_all
    >>> data_to_export = [
    ...     ("data.csv", pd.DataFrame(...)),
    ... ]
    >>> export_all(data_to_export, base_path="data")
"""

from collections.abc import Iterable, Callable
from pathlib import Path
from typing import Type, TypeVar, Any

import pandas as pd

from glytrait.formula import TraitFormula

__all__ = ["export_all"]


def export_all(to_export: Iterable[tuple[str, Any]], base_path: str) -> None:
    """Export all data to the directory given by `base_path`.

    Notes:
        `base_path` will not be created if it does not exist.

    Args:
        to_export: An iterable of tuples (filename, data object).
        base_path: The folder to export data into.
    """
    for filename, data in to_export:
        exporter = _fetch_exporter(data)
        filepath = str(Path(base_path) / filename)
        exporter(data, filepath)


ExportRegistryKey = tuple[Type, bool]
"""key: the object type, value: whether it is a list of objects"""

Exporter = Callable[[Any, str], None]
"""A function that exports an object to a file.
The first argument is the object to export, the second is the filepath.
"""

export_registry: dict[ExportRegistryKey, Exporter] = {}


def register_exporter(data_type: Type, is_list: bool = False):  # type: ignore
    """Decorator for regitering an exporter for a certain type."""
    T = TypeVar("T", bound=Exporter)

    def decorator(exporter_class: T) -> T:
        export_registry[(data_type, is_list)] = exporter_class
        return exporter_class

    return decorator


def _fetch_exporter(single_data: Any) -> Exporter:
    """Factory function for creating an exporter for a data object."""
    is_list = isinstance(single_data, list)
    if is_list:
        try:
            data_type = type(single_data[0])
        except IndexError:
            return dummy_exporter
    else:
        data_type = type(single_data)
    exporter = export_registry.get((data_type, is_list))
    if exporter is not None:
        return exporter
    else:
        raise ValueError(f"Unsupported data type for export: {data_type}.")


def dummy_exporter(data: Any, filepath: str) -> None:
    """Dummy exporter that does nothing."""
    pass


@register_exporter(pd.DataFrame)
def export_dataframe(df: pd.DataFrame, filepath: str) -> None:
    """Export a pandas DataFrame to a CSV file."""
    df.to_csv(filepath, index=True)



"""This module provides functions for calculating meta-properties of glycans.

Functions:
    build_meta_property_table: Build a table of meta-properties for glycans.

Currently, the following meta-properties are supported:
    - type: The type of glycan. (structure only)
    - B: Whether the glycan has bisection. (structure only)
    - nAnt: The number of antennas. (structure only)
    - nF: The number of fucoses.
    - nFc: The number of fucoses on the core. (structure only)
    - nFa: The number of fucoses on the antenna. (structure only)
    - nS: The number of sialic acids.
    - nM: The number of mannoses.
    - nG: The number of galactoses.
    - nN: The number of GlcNAcs.
    - Pl: Whether the glycan has any poly-LacNAc. (structure only)
    - nL: The number of sialic acids with an alpha-2,3 linkage.
    - nE: The number of sialic acids with an alpha-2,6 linkage.
"""

from __future__ import annotations

from enum import Enum, auto
from functools import singledispatch, cache
from typing import Literal, Any, ClassVar, Type

import pandas as pd
from attrs import define
from glypy import Monosaccharide  # type: ignore

from glytrait.data_type import MetaPropertyTable
from glytrait.exception import SiaLinkageError
from glytrait.glycan import Structure, Composition, get_mono_str, GlycanDict

__all__ = [
    "build_meta_property_table",
    "available_meta_properties",
    "get_meta_property",
]


def build_meta_property_table(
    glycans: GlycanDict,
    mode: Literal["composition", "structure"],
    sia_linkage: bool = False,
) -> MetaPropertyTable:
    """Build a table of meta-properties for glycans.

    Args:
        glycans: (GlycanDict): A dict of glycans, with glycan IDs as keys and
            Compositions or Structures as values.
        mode (Literal["composition", "structure"]): The calculation mode.
        sia_linkage (bool, optional): Whether to include the sialic acid linkage
            meta-properties. Defaults to False.

    Returns:
        MetaPropertyTable: The table of meta-properties.
    """
    mp_series_list: list[pd.Series] = []
    for mp_name in available_meta_properties(mode, sia_linkage):
        mp = get_meta_property(mp_name)
        s = mp.calculate_many(glycans)
        mp_series_list.append(s)
    mp_table_df = pd.concat(mp_series_list, axis=1)
    return MetaPropertyTable(mp_table_df)


class GlycanType(Enum):
    """The type of glycan."""

    COMPLEX = auto()
    HIGH_MANNOSE = auto()
    HYBRID = auto()


# ===== Registering meta-properties =====
Glycan = Composition | Structure

_mp_objects: dict[str, MetaProperty] = {}


def register(mp_type: Type[MetaProperty]) -> Type[MetaProperty]:
    """Decorator for registering meta-properties.

    Args:
        mp_type: The meta-property class.

    Returns:
        Type[MetaProperty]: The meta-property class.
    """
    if mp_type.name in _mp_objects:
        raise ValueError(f"Meta-property '{mp_type.name}' already exists")
    _mp_objects[mp_type.name] = mp_type()
    return mp_type


# ===== Get available meta-properties =====
def available_meta_properties(
    mode: Literal["composition", "structure"], sia_linkage: bool
) -> list[str]:
    """Get the available meta-property names.

    Args:
        mode: The calculation mode.
        sia_linkage: Whether to include the sialic acid linkage meta-properties.
            If True, all meta-properties will be included.
            Otherwise, "nE" and "nL" will be excluded.

    Returns:
        list[str]: The available meta-property names.
    """
    to_return: list[str] = []
    for name, mp in _mp_objects.items():
        mode_correct = mp.supported_mode in ("both", mode)  # type: ignore
        if sia_linkage:
            sia_linkage_correct = True
        else:
            sia_linkage_correct = name not in ("nL", "nE")
        if mode_correct and sia_linkage_correct:
            to_return.append(name)
    return to_return


def get_meta_property(name: str) -> MetaProperty:
    """Get a meta-property by its name.

    Args:
        name: The name of the meta-property.

    Returns:
        MetaProperty: The meta-property object.

    Raises:
        KeyError: If the meta-property does not exist.
    """
    return _mp_objects[name]


# ===== Base class for meta-properties =====
@define
class MetaProperty:
    """The base class for meta-properties.

    A meta-property is a basic property of a glycan, such as the number of fucoses,
    the number of sialic acids, etc.
    It is calculated from the glycan structure or composition.

    Besides three class attributes,
    subclasses should also implement the `calculate_one` method for calculating the
    meta-property for a single glycan.
    `calculate_many` is ready to use and should not be overridden.

    For consistency, all subclasses should be named with the suffix "MP".
    e.g. `GlycanTypeMP`, `BisectionMP`, etc.

    Attributes:
        name: The name of the meta-property.
        supported_mode: The supported calculation mode, which can be "composition",
            "structure", or "both".
        return_type: The type of the series returned by `calculate_many`, which can be
            "UInt8", "boolean", or "category".
    """

    name: ClassVar[str]
    supported_mode: ClassVar[Literal["composition", "structure", "both"]]
    return_type: ClassVar[Literal["UInt8", "boolean", "category"]]
    # "UInt8" is used mainly for the number of monosaccharides,
    # which is always non-negative and less than 256.

    def calculate_one(self, glycan: Glycan) -> int | bool | str:
        """Calculate the meta-property for a single glycan.

        Subclasses should always override this method.

        Args:
            glycan: The glycan.

        Returns:
            Any: The value of the meta-property.
        """
        raise NotImplementedError

    def calculate_many(self, glycans: GlycanDict) -> pd.Series:
        """Calculate the meta-property for multiple glycans.

        Args:
            glycans: A dict of glycans, with glycan IDs as keys and
                Compositions or Structures as values.

        Returns:
            pd.Series: A series of the meta-property values, with glycan IDs as the index.
        """
        values = [self.calculate_one(glycan) for glycan in glycans.values()]
        return pd.Series(
            data=values,
            index=pd.Index(glycans.keys()),
            name=self.name,
            dtype=self.return_type,
        )


# ===== Concrete meta-properties =====
@register
@define
class GlycanTypeMP(MetaProperty):
    """The type of glycan."""

    name: ClassVar = "type"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "category"

    def calculate_one(self, glycan: Glycan) -> str:
        """Calculate the type of glycan."""
        return _glycan_type(glycan).name.lower()


@register
@define
class BisectionMP(MetaProperty):
    """Whether the glycan has bisection."""

    name: ClassVar = "B"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "boolean"

    def calculate_one(self, glycan: Glycan) -> bool:
        """Calculate whether the glycan has bisection."""
        return _is_bisecting(glycan)


@register
@define
class CountAntennaMP(MetaProperty):
    """The number of antennas."""

    name: ClassVar = "nAnt"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of antennas."""
        return _count_antenna(glycan)


@register
@define
class CountFucMP(MetaProperty):
    """The number of fucoses."""

    name: ClassVar = "nF"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of fucoses."""
        return _count_fuc(glycan)


@register
@define
class CountCoreFucMP(MetaProperty):
    """The number of fucoses on the core."""

    name: ClassVar = "nFc"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of fucoses on the core."""
        return _count_core_fuc(glycan)


@register
@define
class CountAntennaryFucMP(MetaProperty):
    """The number of fucoses on the antenna."""

    name: ClassVar = "nFa"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of fucoses on the antenna."""
        return _count_fuc(glycan) - _count_core_fuc(glycan)


@register
@define
class CountSiaMP(MetaProperty):
    """The number of sialic acids."""

    name: ClassVar = "nS"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of sialic acids."""
        return _count_sia(glycan)


@register
@define
class CountManMP(MetaProperty):
    """The number of mannoses."""

    name: ClassVar = "nM"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of mannoses."""
        return _count_man(glycan)


@register
@define
class CountGalMP(MetaProperty):
    """The number of galactoses."""

    name: ClassVar = "nG"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of galactoses."""
        return _count_gal(glycan)


@register
@define
class CountGlcNAcMP(MetaProperty):
    """The number of GlcNAcs."""

    name: ClassVar = "nN"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of GlcNAcs."""
        return _count_glcnac(glycan)


@register
@define
class HasPolyLacNAcMP(MetaProperty):
    """Whether the glycan has any poly-LacNAc."""

    name: ClassVar = "Pl"
    supported_mode: ClassVar = "structure"
    return_type: ClassVar = "boolean"

    def calculate_one(self, glycan: Glycan) -> bool:
        """Calculate whether the glycan has any poly-LacNAc."""
        return _has_poly_lacnac(glycan)


@register
@define
class CountA23SiaMP(MetaProperty):
    """The number of sialic acids with an alpha-2,3 linkage."""

    name: ClassVar = "nL"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of sialic acids with an alpha-2,3 linkage."""
        return _count_a23_sia(glycan)


@register
@define
class CountA26SiaMP(MetaProperty):
    """The number of sialic acids with an alpha-2,6 linkage."""

    name: ClassVar = "nE"
    supported_mode: ClassVar = "both"
    return_type: ClassVar = "UInt8"

    def calculate_one(self, glycan: Glycan) -> int:
        """Calculate the number of sialic acids with an alpha-2,6 linkage."""
        return _count_a26_sia(glycan)


# ===== Cacheable functions for calculating meta-properties =====
@cache
def _get_branch_core_man(glycan: Structure) -> tuple[Monosaccharide, Monosaccharide]:
    # Branch core mannose is defined as:
    #
    # X - Man  <- This one
    #        \
    #         Man - Glc2NAc - Glc2NAc
    #        /
    # X - Man  <- And this one
    bft_iter = glycan.breadth_first_traversal(skip=["Fuc"])
    for i in range(3):
        next(bft_iter)
    while True:
        node1 = next(bft_iter)
        if get_mono_str(node1) == "Man":
            break
    while True:
        node2 = next(bft_iter)
        if get_mono_str(node2) == "Man":
            break
    return node1, node2


@cache
def _glycan_type(glycan: Structure) -> GlycanType:
    """Decide whether a glycan is a 'complex', 'hybrid', or 'high-mannose' type."""
    # N-glycan core is defined as the complex type.
    if glycan.composition == {"Glc2NAc": 2, "Man": 3}:
        # This type of glycan is actually pausimannose type.
        # However, it is currently regarded as a complex type right now.
        return GlycanType.COMPLEX

    # Bisecting could only be found in complex type.
    if _is_bisecting(glycan):
        return GlycanType.COMPLEX

    # If the glycan is not core, and it only has 2 "GlcNAc", it is high-mannose.
    if glycan.composition["Glc2NAc"] == 2:
        return GlycanType.HIGH_MANNOSE

    # If the glycan is mono-antennary and not high-monnose, it is complex.
    node1, node2 = _get_branch_core_man(glycan)
    if any((len(node1.links) == 1, len(node2.links) == 1)):
        return GlycanType.COMPLEX

    # Then, if it has 3 "Glc2NAc", it must be hybrid.
    if glycan.composition["Glc2NAc"] == 3:
        return GlycanType.HYBRID

    # All rest cases are complex.
    return GlycanType.COMPLEX


@cache
def _is_bisecting(glycan: Structure) -> bool:
    """Decide whether a glycan has bisection."""
    bft_iter = glycan.breadth_first_traversal(skip=["Fuc"])
    for i in range(2):
        next(bft_iter)
    next_node = next(bft_iter)
    return len(next_node.links) == 4


@cache
def _count_antenna(glycan: Structure) -> int:
    """Count the number of branches in a glycan."""
    if _glycan_type(glycan) != GlycanType.COMPLEX:
        return 0
    node1, node2 = _get_branch_core_man(glycan)
    return len(node1.links) + len(node2.links) - 2


@singledispatch
def _count_fuc(glycan) -> int:
    """Count the number of fucoses."""
    raise TypeError


@_count_fuc.register
@cache
def _(glycan: Structure) -> int:
    return glycan.composition.get("Fuc", 0)


@_count_fuc.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("F", 0)


@cache
def _get_core_ids(glycan: Structure) -> list[int]:
    """Get the IDs of the monosaccharides of the core."""
    should_be = ["Glc2NAc", "Glc2NAc", "Man", "Man", "Man"]
    cores: list[int] = []
    for node in glycan.breadth_first_traversal(skip=["Fuc"]):
        # The first two monosaacharides could only be "Glc2NAc".
        # However, when the glycan is bisecting, the rest of the three monosaccharides
        # might not all be "Man". So we look for the monosaacharides in the order of
        # `should_be`, and skip the ones that are not in the order.
        if get_mono_str(node) == should_be[len(cores)]:
            cores.append(node.id)
        if len(cores) == 5:
            break
    return cores


@cache
def _count_core_fuc(glycan: Structure) -> int:
    cores = _get_core_ids(glycan)
    n = 0
    for node in glycan.breadth_first_traversal():
        # node.parents()[0] is the nearest parent, and is a tuple of (Link, Monosaccharide)
        # node.parents()[0][1] is the Monosaccharide
        if get_mono_str(node) == "Fuc" and node.parents()[0][1].id in cores:
            n = n + 1
    return n


@singledispatch
def _count_sia(glycan) -> int:
    """Count the number of sialic acids."""
    raise TypeError


@_count_sia.register
@cache
def _(glycan: Structure) -> int:
    return glycan.composition.get("Neu5Ac", 0) + glycan.composition.get("Neu5Gc", 0)


@_count_sia.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("S", 0) + glycan.get("E", 0) + glycan.get("L", 0)


@singledispatch
def _count_gal(glycan) -> int:
    """Count the number of Gals."""
    raise TypeError


@_count_gal.register
@cache
def _(glycan: Structure) -> int:
    return glycan.composition.get("Gal", 0)


@_count_gal.register
@cache
def _(glycan: Composition) -> int:
    n_H = glycan.get("H", 0)
    n_N = glycan.get("N", 0)
    if n_H >= 4 and n_N >= n_H - 1:
        return n_H - 3
    return 0


@singledispatch
def _count_man(glycan) -> int:
    """Count the number of mannoses."""
    raise TypeError


@_count_man.register
@cache
def _(glycan: Structure) -> int:
    return glycan.composition.get("Man", 0)


@_count_man.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("H", 0) - _count_gal(glycan)


@singledispatch
def _count_glcnac(glycan) -> int:
    """Count the number of GlcNAcs."""
    raise TypeError


@_count_glcnac.register
@cache
def _(glycan: Structure) -> int:
    return glycan.composition.get("Glc2NAc", 0)


@_count_glcnac.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("N", 0)


@cache
def _has_poly_lacnac(glycan: Structure) -> bool:
    """Decide whether a glycan has any poly-LacNAc."""
    # Iterate all Gal residues.
    # If a Gal residue has a GlcNAc child, and the GlcNAc residue has a Gal child,
    # then the glycan has poly-LacNAc.
    gals = glycan.breadth_first_traversal(only=["Gal"])
    for gal in gals:
        for _, child in gal.children():
            if get_mono_str(child) == "Glc2NAc":
                children_of_glcnac = child.children()
                for _, child_of_glcnac in children_of_glcnac:
                    if get_mono_str(child_of_glcnac) == "Gal":
                        return True
    return False


@singledispatch
def _count_a23_sia(glycan) -> int:
    """Count the number of sialic acids with an alpha-2,3 linkage."""
    raise TypeError


@_count_a23_sia.register
@cache
def _(glycan: Structure) -> int:
    n = 0
    for node in glycan.breadth_first_traversal():
        if get_mono_str(node) in ("Neu5Ac", "Neu5Gc"):
            if node.links[2][0].parent_position == -1:
                raise SiaLinkageError("Sialic acid linkage not specified")
            elif node.links[2][0].parent_position == 3:
                n = n + 1
    return n


@_count_a23_sia.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("L", 0)


@singledispatch
def _count_a26_sia(glycan) -> int:
    """Count the number of sialic acids with an alpha-2,6 linkage."""
    raise TypeError


@_count_a26_sia.register
@cache
def _(glycan: Structure) -> int:
    n = 0
    for node in glycan.breadth_first_traversal():
        if get_mono_str(node) in ("Neu5Ac", "Neu5Gc"):
            if node.links[2][0].parent_position == -1:
                raise SiaLinkageError("Sialic acid linkage not specified")
            elif node.links[2][0].parent_position == 6:
                n = n + 1
    return n


@_count_a26_sia.register
@cache
def _(glycan: Composition) -> int:
    return glycan.get("E", 0)



"""This module provides the functions to calculate and filter the derived traits.

This module implements only one function: `calcu_derived_trait`.
"""

import pandas as pd

from glytrait.formula import TraitFormula
from glytrait.data_type import AbundanceTable, MetaPropertyTable, DerivedTraitTable

__all__ = ["calcu_derived_trait"]


def calcu_derived_trait(
    abund_df: AbundanceTable,
    meta_prop_df: MetaPropertyTable,
    formulas: list[TraitFormula],
) -> DerivedTraitTable:
    """Calculate the derived trait values.

    Args:
        abund_df (AbundanceTable): The abundance table, with samples as index and Compositions
            as columns.
        meta_prop_df (MetaPropertyTable): The table of meta properties generated by
            `build_meta_property_table`.
        formulas (list[TraitFormula]): The trait formulas.

    Returns:
        DerivedTraitTable: The trait values, with samples as index and trait names as columns.
    """
    # The meta-property table the formulas use for initialization must have
    # the same order of index as the abundance table's columns (glycans).
    # `TraitFormula.calcu_trait` will raise an assertion error if the orders
    # of the glycans are different.
    # # (See `TraitFormula`)
    meta_prop_df_ordered = MetaPropertyTable(meta_prop_df.reindex(abund_df.columns))
    trait_series: list[pd.Series] = []
    for formula in formulas:
        formula.initialize(meta_prop_df_ordered)
        trait_series.append(formula.calcu_trait(abund_df))
    derived_trait_df = pd.concat(trait_series, axis=1)
    derived_trait_df = derived_trait_df.round(6)
    return DerivedTraitTable(derived_trait_df)



"""This module defines the `TraitFormula` class, and other related functions.

The `TraitFormula` class is used to represent a trait formula.
It is the core of calculating derived traits.

Classes:
    TraitFormula: The trait formula.

Functions:
    load_formulas: Load all trait formulas from default formula files and custom formulas files.
    save_trait_formulas_tepmlate: Save a template of trait formulas to a file.
    save_builtin_formula: Save a built-in trait formulas to a file.
"""

from __future__ import annotations

import itertools
import functools
import re
from collections.abc import Callable, Iterator
from importlib.resources import files, as_file
from pathlib import Path
from typing import Literal, Optional, Generator, Type, Protocol

import numpy as np
import pandas as pd
import attrs
from attrs import field, define

from glytrait.data_type import AbundanceTable, MetaPropertyTable

__all__ = [
    "TraitFormula",
    "load_formulas",
    "load_default_formulas",
    "save_builtin_formula",
]

from glytrait.exception import GlyTraitError

default_struc_formula_file = files("glytrait.resources").joinpath("struc_formula.txt")
default_comp_formula_file = files("glytrait.resources").joinpath("comp_formula.txt")


class FormulaError(GlyTraitError):
    """Raised if a formula is invalid."""


class FormulaFileError(GlyTraitError):
    """Raised if a formula file is in wrong format."""


class FormulaParseError(FormulaError):
    """Raised when failing parsing a formula expression."""

    def __init__(self, expr: str, reason: str = ""):
        super().__init__()
        self.expr = expr
        self.reason = reason

    def __str__(self):
        return f"Invalid expression: {self.expr}. {self.reason}"


class FormulaTermParseError(FormulaParseError):
    """Raised when failing parsing a formula term expression."""


class FormulaCalculationError(FormulaError):
    """Raised when failing to calculate a derived trait."""


class FormulaTermCalculationError(FormulaCalculationError):
    """Raised when calling a formula term fails."""

    def __init__(self, term: str):
        super().__init__()
        self.term = term


class MissingMetaPropertyError(FormulaTermCalculationError):
    """Raised when a meta-property is missing in the meta-property table."""

    def __init__(self, term: str, mp: str):
        super().__init__(term)
        self.mp = mp

    def __str__(self):
        return f"Meta-property '{self.mp}' is missing in the meta-property table."


class MetaPropertyTypeError(FormulaTermCalculationError):
    """Raised when the dtype of a meta-property is not supported."""

    def __init__(self, term: str, mp: str, dtype: str, reason: str = ""):
        super().__init__(term)
        self.mp = mp
        self.dtype = dtype
        self.reason = reason

    def __str__(self):
        return (
            f"Meta-property '{self.mp}' has an unsupported dtype: "
            f"'{self.dtype}'. {self.reason}"
        )


class FormulaNotInitializedError(FormulaError):
    """Raised when a formula is not initialized."""


class FormulaTerm(Protocol):
    """The protocol for a formula term."""

    def __call__(self, meta_property_table: MetaPropertyTable) -> pd.Series: ...

    @property
    def expr(self) -> str: ...


class FormulaTermWithParser(FormulaTerm, Protocol):

    @classmethod
    def from_expr(cls, expr: str) -> FormulaTerm: ...


_terms_with_parser: list[type[FormulaTermWithParser]] = []


def _register_term(cls):
    """A decorator to register a formula term class."""
    _terms_with_parser.append(cls)
    return cls


def validate_parentheses(*, must_have: bool = False):
    """Validate the parentheses in the `expr` argument.

    This decorator is intended to be used on `FormulaTerm.from_expr`.

    Args:
        must_have: Whether the expression must have the parentheses.
            Default to False.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(cls, expr: str):
            if must_have:
                if not (expr.startswith("(") and expr.endswith(")")):
                    raise FormulaTermParseError(
                        expr, "This term must have parentheses."
                    )
            if expr.startswith("(") and not expr.endswith(")"):
                raise FormulaTermParseError(expr, "Missing ')'.")
            if expr.endswith(")") and not expr.startswith("("):
                raise FormulaTermParseError(expr, "Missing '('.")
            if "(" in expr.strip("()") or ")" in expr.strip("()"):
                raise FormulaTermParseError(expr, "Too many parentheses.")
            return func(cls, expr)

        return wrapper

    return decorator


def remove_parentheses(func):
    """Remove the parentheses in the `expr` argument.

    This decorator is intended to be used on `FormulaTerm.from_expr`.

    Notes:
        This decorator should be used inside `validate_parentheses`.
    """

    @functools.wraps(func)
    def wrapper(cls, expr: str):
        expr = expr.strip("()")
        return func(cls, expr)

    return wrapper


@_register_term
@define
class ConstantTerm:
    """Return a series with all values being constant.

    Args:
        value: The constant value. It should be a float greater than 0.
    """

    value: float = field(validator=attrs.validators.gt(0))

    def __call__(self, meta_property_table: MetaPropertyTable) -> pd.Series:
        """Calculate the term."""
        return pd.Series(
            data=self.value,
            index=meta_property_table.index,
            name=self.expr,
            dtype="Float32",
        )

    @property
    def expr(self) -> str:
        """The expression of the term."""
        return str(self.value)

    @classmethod
    @validate_parentheses(must_have=False)
    @remove_parentheses
    def from_expr(cls, expr: str) -> ConstantTerm:
        """Create a formula term from an expression.

        Args:
            expr: The expression of the term.

        Returns:
            The formula term.

        Raises:
            FormulaTermParseError: If the expression is invalid.
        """
        try:
            value = float(expr)
        except ValueError as e:
            reason = "Could not be parsed into a ConstantTerm."
            raise FormulaTermParseError(expr, reason) from e
        return cls(value=value)


@_register_term
@define
class NumericalTerm:
    """Return the values of a numerical meta-property.

    This term simply returns a column of the meta-property table,
    as a Series with the same index as the meta-property table.

    Note that the dtype of the meta-property should be UInt8.

    Args:
        meta_property: The numerical meta property.
    """

    meta_property: str = field()

    def __call__(self, meta_property_table: MetaPropertyTable) -> pd.Series:
        """Calculate the term.

        The return value is a Series with the same index as the meta-property table.

        Args:
            meta_property_table: The table of meta properties.

        Returns:
            pd.Series: The values of the meta-property,
            with the same index as the meta-property table.

        Raises:
            MissingMetaPropertyError: If the meta-property is not in the meta-property table.
            MetaPropertyTypeError: If the dtype of the meta-property is not UInt8.
        """
        try:
            mp_s = meta_property_table[self.meta_property]
        except KeyError as e:
            raise MissingMetaPropertyError(self.expr, self.meta_property) from e

        if mp_s.dtype == "category":
            reason = "NumericalTerm can't be working with categorical meta-properties."
            raise MetaPropertyTypeError(self.meta_property, str(mp_s.dtype), reason)

        return mp_s.astype("UInt8")

    @property
    def expr(self) -> str:
        """The expression of the term."""
        return self.meta_property

    @classmethod
    @validate_parentheses(must_have=False)
    @remove_parentheses
    def from_expr(cls, expr: str) -> NumericalTerm:
        """Create a formula term from an expression."""
        meets_condition = bool(re.fullmatch(r"\w+", expr)) and not expr.isdigit()
        if not meets_condition:
            reason = "Could not be parsed into a NumericalTerm."
            raise FormulaTermParseError(expr, reason)
        return cls(meta_property=expr)


OPERATORS = {"==", "!=", ">", ">=", "<", "<="}


@_register_term
@define
class CompareTerm:
    """Compare the value of a meta-property with a given value.

    The validity of `meta_property` will not be checked here.

    Args:
        meta_property: The meta property to compare.
        operator: The comparison operator.
        value: The value to compare with.
    """

    meta_property: str = field()
    operator: Literal["==", "!=", ">", ">=", "<", "<="] = field()
    value: float | bool | str = field()

    def __call__(self, meta_property_table: MetaPropertyTable) -> pd.Series:
        """Calculate the term.

        Args:
            meta_property_table: The table of meta properties.

        Returns:
            pd.Series: A boolean Series with the same index as the meta-property table.

        Raises:
            MissingMetaPropertyError: If the meta-property is not in the meta-property table.
            MetaPropertyTypeError: If the dtype of the meta-property is not supported.
        """
        try:
            mp_s = meta_property_table[self.meta_property]
        except KeyError as e:
            raise MissingMetaPropertyError(self.expr, self.meta_property) from e

        condition_1 = mp_s.dtype == "boolean" or mp_s.dtype == "category"
        condition_2 = self.operator in {">", ">=", "<", "<="}
        if condition_1 and condition_2:
            reason = (
                f"Cannot use '{self.operator}' with '{mp_s.dtype}' meta properties."
            )
            raise MetaPropertyTypeError(
                self.expr, self.meta_property, str(mp_s.dtype), reason
            )

        if isinstance(self.value, str):
            expr = f"mp_s {self.operator} '{self.value}'"
        else:
            expr = f"mp_s {self.operator} {self.value}"
        result_s = eval(expr)

        result_s.name = self.expr
        result_s = result_s.astype("UInt8")
        return result_s

    @property
    def expr(self) -> str:
        """The expression of the term."""
        if isinstance(self.value, str):
            return f"{self.meta_property} {self.operator} '{self.value}'"
        else:
            return f"{self.meta_property} {self.operator} {self.value}"

    @staticmethod
    def _valid_operator(expr: str) -> bool:
        """Check if the expression has a valid operator."""
        for op in OPERATORS:
            if op in expr:
                return True
        return False

    @classmethod
    @validate_parentheses(must_have=False)
    @remove_parentheses
    def from_expr(cls, expr: str) -> CompareTerm:
        """Create a formula term from an expression."""
        if not cls._valid_operator(expr):
            raise FormulaTermParseError(expr, "Conditions not met.")

        meta_property_p = r"(\w+)"
        operator_p = r"(==|!=|>|>=|<|<=)"
        value_p = r"""(\d+|True|False|'[a-zA-Z-_]*'|"[a-zA-Z-_]*")"""
        total_p = rf"{meta_property_p}\s*{operator_p}\s*{value_p}"
        match = re.fullmatch(total_p, expr)
        if match is None:
            raise FormulaTermParseError(expr, "Unknown error.")
        meta_property, operator, value = match.groups()

        if value.isdigit():
            value = int(value)
        elif value == "True":
            value = True
        elif value == "False":
            value = False
        else:
            value = value.strip("'")
            value = value.strip('"')

        return cls(  # type: ignore
            meta_property=meta_property, operator=operator, value=value  # type: ignore
        )


def _is_sia_linkage_term(term: FormulaTerm) -> bool:
    """Check if a formula term is related to sia-linkage."""
    return "nE" in term.expr or "nL" in term.expr


@define
class DivisionTermWrapper:
    """Change a term into a division term.

    This is a wrapper class for a formula term.
    If calling the original term results a vector of `a`,
    this wrapper will return a vector of `1/a`.
    Specially, if any element of `a` is zero,
    the corresponding element of the result will also be zero.

    Notes:
        The original term should not be a `CompareTerm`.

    Args:
        term: The original term.
    """

    term: FormulaTerm = field()

    @term.validator
    def _validate_term(self, attribute, value):
        if value.__class__ is CompareTerm:
            raise ValueError("The original term should not be a CompareTerm.")

    def __call__(self, meta_property_table: MetaPropertyTable) -> pd.Series:
        """Calculate the term.

        Raises:
            FormulaTermCalculationError: If the original term calculation fails.
        """
        series = self.term(meta_property_table)
        array = np.array(series.values, dtype=float)
        return pd.Series(
            np.divide(1, array, where=(array != 0)),
            dtype="Float32",
            name=self.expr,
            index=series.index,
        )

    @property
    def expr(self) -> str:
        """The expression of the term."""
        return f"/ ({self.term.expr})"


FormulaParserType = Callable[[str], "TraitFormula"]


@define
class TraitFormula:
    """The trait formula.

    This class is used for calculating a derived trait in a vectorized manner.

    Each formula has a name, a description, and two lists of formula terms:
    numerators and denominators.
    The formula terms are instances of `FormulaTerm`.

    To create a `TraitFormula` instance, initialize it with the name, the description,
    and the lists of formula terms,
    or use the class method `from_expr` to parse the expression of the formula.

    To calculate the derived trait values,
    first call the `initialize` method on a meta-property table,
    then call the `calcu_trait` method on an abundance table.

    Attributes:
        name: The name of the formula.
        numerators: The list of formula terms in the numerator.
        denominators: The list of formula terms in the denominator.
        sia_linkage: Whether the formula is related to sia-linkage.

    Examples:
        >>> expr = "Trait = [A * B] / [C * D]"
        >>> formula = TraitFormula.from_expr(expr)
        >>> formula.name
        'Trait'
        >>> formula.initialize(meta_property_table)
        >>> trait_values = formula.calcu_trait(abundance_table)
    """

    name: str = field()
    numerators: list[FormulaTerm] = field(validator=attrs.validators.min_len(1))
    denominators: list[FormulaTerm] = field(validator=attrs.validators.min_len(1))

    _sia_linkage: bool = field(init=False)
    _initialized: bool = field(init=False, default=False)
    _numerator_s: pd.Series = field(init=False, default=None)
    _denominator_s: pd.Series = field(init=False, default=None)

    def __attrs_post_init__(self):
        self._sia_linkage = self._init_sia_linkage()

    def _init_sia_linkage(self) -> bool:
        """Check if the formula is related to sia-linkage."""
        term_iter = itertools.chain(self.numerators, self.denominators)
        return any(_is_sia_linkage_term(term) for term in term_iter)

    @classmethod
    def from_expr(
        cls,
        expr: str,
        *,
        parser: Optional[FormulaParserType] = None,  # Dependency injection
    ) -> TraitFormula:
        """Create a formula from an expression.

        Args:
            expr: The expression of the formula.
            parser: The parser to parse the formula expression.
                If None, the default parser will be used.
                Defaults to None.

        Returns:
            The formula.

        Raises:
            FormulaParseError: If the expression is invalid.
        """
        if parser is None:
            parser = FormulaParser()
        return parser(expr)

    @property
    def sia_linkage(self) -> bool:
        """Whether the formula is related to sia-linkage."""
        return self._sia_linkage

    def initialize(self, meta_property_table: MetaPropertyTable) -> None:
        """Initialize the trait formula.

        Args:
            meta_property_table: The table of meta properties.

        Raises:
            FormulaCalculationError: If the initialization fails.
        """
        self._numerator_s = self._initialize(meta_property_table, self.numerators)
        self._denominator_s = self._initialize(meta_property_table, self.denominators)
        self._initialized = True

    @staticmethod
    def _initialize(
        meta_property_table: MetaPropertyTable, terms: list[FormulaTerm]
    ) -> pd.Series:
        """Initialize the numerator or denominator of the formula."""
        try:
            series_list = [term(meta_property_table) for term in terms]
        except FormulaTermCalculationError as e:
            msg = f"Failed to calculate term: {e.term}. {str(e)}"
            raise FormulaCalculationError(msg) from e
        return pd.concat(series_list, axis=1).prod(axis=1)

    def calcu_trait(self, abundance_table: AbundanceTable) -> pd.Series:
        """Calculate the trait.

        Args:
            abundance_table (AbundanceTable): The glycan abundance table,
                with samples as index, and glycans as columns.

        Returns:
            pd.Series: An array of trait values for each sample.

        Raises:
            FormulaNotInitializedError: If the formula is not initialized.
        """
        if not self._initialized:
            raise FormulaNotInitializedError()
        pd.testing.assert_index_equal(abundance_table.columns, self._numerator_s.index)
        pd.testing.assert_index_equal(
            abundance_table.columns, self._denominator_s.index
        )

        numerator_array = np.array(self._numerator_s.values, dtype=float)
        denominator_array = np.array(self._denominator_s.values, dtype=float)
        numerator = abundance_table.values @ numerator_array
        denominator = abundance_table.values @ denominator_array
        denominator[denominator == 0] = np.nan
        values = numerator / denominator
        return pd.Series(
            values, index=abundance_table.index, name=self.name, dtype=float
        )


@define
class FormulaParser:
    """Parser of the trait formula expressions.

    The basic format of a formula expression is:

    '<name> = [<numerators>] / [<denominators>]'
    '<name> = [<numerators>] // [<denominators>]'

    In the second format, the numerators will be updated by the denominators.
    For example, 'A = [B] // [C]' is the same as 'A = [B * C] / [C]'.

    Numerators and denominators are separated by '*' or '/'.
    For example, the following expressions are all valid:

    - 'A = [B * C] / [D]'
    - 'A = [B * C / D] // [E]'  # The same as 'A = [B * C * E / D] / [E]'

    Any specific term should be a valid term expression.
    See the concrete term classes (subclasses of `FormulaTerm`) for more details.
    However, '*' and '/' are not allowed in the parentheses.

    More examples of valid expressions:

    - 'A = [B * C] / [D]'
    - 'A = [B * C / D] // [E]'
    - 'A = [(B == 1) * C] // [D]'

    Examples:
        >>> parser = FormulaParser()
        >>> formula1 = parser("expr1")  # As callable
        >>> formula2 = parser.parse("expr2")  # Or use the `parse` method
    """

    _available_terms: list[Type[FormulaTermWithParser]] = field(
        kw_only=True, default=None
    )
    _formula_factory: Type[TraitFormula] = field(kw_only=True, default=TraitFormula)

    def __attrs_post_init__(self):
        if self._available_terms is None:
            self._available_terms = list(_terms_with_parser)

    def __call__(self, expr: str) -> TraitFormula:
        return self.parse(expr)

    def parse(self, expr: str) -> TraitFormula:
        """Parse the formula expression.

        Args:
            expr: The formula expression.

        Returns:
            The `TraitFormula` object.

        Raises:
            FormulaParseError: If the expression is invalid.
        """
        name, numerator_expr, spliter, denominator_expr = self._split_expr(expr)
        numerators = self._parse_terms_expr(numerator_expr)
        denominators = self._parse_terms_expr(denominator_expr)
        if spliter == "//":
            numerators.extend(denominators)
        return self._formula_factory(name, numerators, denominators)

    @staticmethod
    def _split_expr(expr: str) -> tuple[str, str, str, str]:
        """Split the formula expression into four parts:

        - The name of the formula.
        - The numerator expression of the formula.
        - The slash or double-slash.
        - The denominator expression of the formula.
        """
        expr = expr.strip()
        if " = " not in expr:
            raise FormulaParseError(expr, "Missing ' = ' (spaces are important).")

        try:
            name, expr_after_name = expr.split(" = ")
        except ValueError:
            raise FormulaParseError(expr, "Misuse of '=' for '=='.")
        name = name.strip()
        expr_after_name = expr_after_name.strip()

        pattern = r"\[(.*?)\] (/{1,2}) \[(.*?)\]"
        match = re.fullmatch(pattern, expr_after_name)
        if match is None:
            raise FormulaParseError(expr, "Invalid format.")

        numerator_expr, spliter, denominator_expr = match.groups()
        numerator_expr = numerator_expr.strip()
        denominator_expr = denominator_expr.strip()
        return name, numerator_expr, spliter, denominator_expr

    def _parse_terms_expr(self, expr: str) -> list[FormulaTerm]:
        """Parse the numerator or denominator expression."""
        terms: list[FormulaTerm] = []
        for s, t in self._split_terms(expr):
            try:
                term = self._parse_term(t)
            except FormulaTermParseError as e:
                raise FormulaParseError(expr, str(e)) from e
            if s == "*":
                terms.append(term)
            else:  # s == "/"
                terms.append(DivisionTermWrapper(term))
        return terms

    @staticmethod
    def _split_terms(expr: str) -> list[tuple[str, str]]:
        """Split the numerator or denominator expression into symbols and terms."""
        expr = "* " + expr
        pattern = r"(\*|/)([^\*/]*)"
        matches = re.findall(pattern, expr)
        return [(s, t.strip()) for s, t in matches]

    def _parse_term(self, expr: str) -> FormulaTerm:
        """Parse a term expression."""
        for term_cls in self._available_terms:
            try:
                return term_cls.from_expr(expr)  # Let the exception pass through
            except FormulaTermParseError:
                continue
        raise FormulaTermParseError(expr, "Does not belong to any term class.")


@define
class FormulaFileParser:
    """Parser for a formula file."""

    expr_parser: FormulaParserType = field(kw_only=True, default=FormulaParser())

    def parse(self, filepath: str) -> Iterator[TraitFormula]:
        """Parse the formula file and return a list of `TraitFormula` objects."""
        for description, expression in self._deconvolute_formula_file(filepath):
            yield TraitFormula.from_expr(expression, parser=self.expr_parser)

    @staticmethod
    def _deconvolute_formula_file(formula_file: str) -> Iterator[tuple[str, str]]:
        """A generator that yields the formula description and the formula expression.

        Args:
            formula_file (str): The path of the formula file.

        Yields:
            tuple[str, str]: The formula description and the formula expression.

        Raises:
            FormulaFileError: If the formula file is in a wrong format.
        """
        with open(formula_file, "r", encoding="utf8") as f:
            description, expression = None, None

            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue

                if line.startswith("@"):
                    if description is not None:
                        msg = f"No expression follows description '{description}'."
                        raise FormulaFileError(msg)
                    description = line[1:].strip()

                elif line.startswith("$"):
                    expression = line[1:].strip()
                    if description is None:
                        msg = f"No description before expression '{expression}'."
                        raise FormulaFileError(msg)
                    yield description, expression
                    description, expression = None, None

        if description is not None:
            raise FormulaFileError(
                f"No expression follows description '{description}'."
            )


def load_formulas(
    file: str,
    sia_linkage: bool = False,
) -> list[TraitFormula]:
    """Load formulas from a formula file.

    Args:
        file: The path of the formula file.
        sia_linkage: Whether to include formulas with sia-linkage
            meta-properties. Defaults to False.

    Returns:
        list[TraitFormula]: The formulas.

    Raises:
        FormulaFileError: If the formula file is in a wrong format.
        FormulaParseError: If a formula string cannot be parsed.
    """
    formulas = [f for f in FormulaFileParser().parse(file)]
    if not sia_linkage:
        formulas = [f for f in formulas if not f.sia_linkage]
    return formulas


def load_default_formulas(
    mode: Literal["structure", "composition"],
    sia_linkage: bool = False,
) -> list[TraitFormula]:
    """Load the default formulas.

    Args:
        mode: The type of the formulas.
        sia_linkage: Whether to include formulas with sia-linkage
            meta-properties. Defaults to False.

    Returns:
        list[TraitFormula]: The formulas.
    """
    if mode == "composition":
        file_traversable = default_comp_formula_file
    elif mode == "structure":
        file_traversable = default_struc_formula_file
    else:
        raise ValueError("Invalid formula type.")
    with as_file(file_traversable) as file:
        return load_formulas(str(file), sia_linkage=sia_linkage)


def save_builtin_formula(dirpath: str | Path) -> None:
    """Copy the builtin formula file to the given path.

    Args:
        dirpath (str): The path to save the built-in formula file.
    """
    Path(dirpath).mkdir(parents=True, exist_ok=True)
    struc_file = Path(dirpath) / "struc_builtin_formulas.txt"
    comp_file = Path(dirpath) / "comp_builtin_formulas.txt"
    struc_content = default_struc_formula_file.open("r").read()
    comp_content = default_comp_formula_file.open("r").read()
    with open(struc_file, "w", encoding="utf8") as f:
        f.write(struc_content)
    with open(comp_file, "w", encoding="utf8") as f:
        f.write(comp_content)
